/* Tree-dumping functionality for intermediate representation.
   Copyright (C) 1999, 2000 Free Software Foundation, Inc.
   Written by Mark Mitchell <mark@codesourcery.com>

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.  */

/* T. Dean, February 2001
 * - changed to dump in our own format.
 */

#include "config.h"
#include "system.h"
#include "tree.h"
#include "c-common.h"
#include "splay-tree.h"
#include "diagnostic.h"
#include "toplev.h"
#include "c-dump.h"
#include "gd_graph.h"

/* A callback function used dump language-specific parts of tree
   nodes.  Returns non-zero if it does not want the usual dumping of
   the second argument.  */

dump_tree_fn lang_dump_tree;

gd_graph the_graph;

static unsigned int queue PARAMS ((dump_info_p, tree, int));
static void dump_index PARAMS ((dump_info_p, unsigned int));
static void dequeue_and_dump PARAMS ((dump_info_p));
static void dump_new_line PARAMS ((dump_info_p));
static void dump_maybe_newline PARAMS ((dump_info_p));
static void dump_node PARAMS ((tree, FILE *));

/* Add T to the end of the queue of nodes to dump.  Returns the index
   assigned to T.  */

static unsigned int
queue (di, t, flags)
     dump_info_p di;
     tree t;
     int flags;
{
  dump_queue_p dq;
  dump_node_info_p dni;
  unsigned int index;

  /* Assign the next available index to T.  */
  index = ++di->index;

  /* Obtain a new queue node.  */
  if (di->free_list)
    {
      dq = di->free_list;
      di->free_list = dq->next;
    }
  else
    dq = (dump_queue_p) xmalloc (sizeof (struct dump_queue));

  /* Create a new entry in the splay-tree.  */
  dni = (dump_node_info_p) xmalloc (sizeof (struct dump_node_info));
  dni->index = index;
  dni->binfo_p = ((flags & DUMP_BINFO) != 0);
  dq->node = splay_tree_insert (di->nodes, (splay_tree_key) t, 
				(splay_tree_value) dni);

  /* Add it to the end of the queue.  */
  dq->next = 0;
  if (!di->queue_end)
    di->queue = dq;
  else
    di->queue_end->next = dq;
  di->queue_end = dq;

  /* Return the index.  */
  return index;
}

static void
dump_index (di, index)
     dump_info_p di;
     unsigned int index;
{
  fprintf (di->stream, "@%-6u ", index);
  di->column += 8;
}

/* If T has not already been output, queue it for subsequent output.
   FIELD is a string to print before printing the index.  Then, the
   index of T is printed.  */

void
queue_and_dump_index (di, field, t, flags)
     dump_info_p di;
     const char *field;
     tree t;
     int flags;
{
  unsigned int index;
  splay_tree_node n;

  /* If there's no node, just return.  This makes for fewer checks in
     our callers.  */
  if (!t)
    return;

  /* See if we've already queued or dumped this node.  */
  n = splay_tree_lookup (di->nodes, (splay_tree_key) t);
  if (n)
    index = ((dump_node_info_p) n->value)->index;
  else
    /* If we haven't, add it to the queue.  */
    index = queue (di, t, flags);

  /* Print the index of the node.  */
  dump_maybe_newline (di);
  fprintf (di->stream, "%-4s: ", field);
  di->column += 6;
  dump_index (di, index);
}

unsigned long
queue_and_return_index (di, t, flags)
     dump_info_p di;
     tree t;
     int flags;
{
  unsigned int index;
  splay_tree_node n;

  /* If there's no node, just return.  This makes for fewer checks in
     our callers.  */
  if (!t)
    return 0;

  /* See if we've already queued or dumped this node.  */
  n = splay_tree_lookup (di->nodes, (splay_tree_key) t);
  if (n)
    index = ((dump_node_info_p) n->value)->index;
  else
    /* If we haven't, add it to the queue.  */
    index = queue (di, t, flags);

  /* return the index of the node.  */
  return index;
}

/* Dump the type of T.  */

void
queue_and_dump_type (di, t)
     dump_info_p di;
     tree t;
{
  queue_and_dump_index (di, "type", TREE_TYPE (t), DUMP_NONE);
}

/* Insert a new line in the dump output, and indent to an appropriate
   place to start printing more fields.  */

static void
dump_new_line (di)
     dump_info_p di;
{
  fprintf (di->stream, "\n%25s", "");
  di->column = 25;
}

/* If necessary, insert a new line.  */

static void
dump_maybe_newline (di)
     dump_info_p di;
{
  /* See if we need a new line.  */
  if (di->column > 53)
    dump_new_line (di);
  /* See if we need any padding.  */
  else if ((di->column - 25) % 14 != 0)
    {
      fprintf (di->stream, "%*s", 14 - ((di->column - 25) % 14), "");
      di->column += 14 - (di->column - 25) % 14;
    }
}

/* Dump I using FIELD to identity it.  */

void
dump_int (di, field, i)
     dump_info_p di;
     const char *field;
     int i;
{
  dump_maybe_newline (di);
  fprintf (di->stream, "%-4s: %-7d ", field, i);
  di->column += 14;
}

/* Dump the string S.  */

void
dump_string (di, string)
     dump_info_p di;
     const char *string;
{
  dump_maybe_newline (di);
  fprintf (di->stream, "%-13s ", string);
  if (strlen (string) > 13)
    di->column += strlen (string) + 1;
  else
    di->column += 14;
}

/* Dump information common to statements from STMT.  */

void
dump_stmt (di, t)
     dump_info_p di;
     tree t;
{
  dump_int (di, "line", STMT_LINENO (t));
}

/* Dump the next statement after STMT.  */

void
dump_next_stmt (di, t)
     dump_info_p di;
     tree t;
{
  dump_child ("next", TREE_CHAIN (t));
}

/* Dump the next node in the queue.  */

static void 
dequeue_and_dump (di)
     dump_info_p di;
{
  dump_queue_p dq;
  splay_tree_node stn;
  dump_node_info_p dni;
  tree t;
  unsigned int index;
  enum tree_code code;
  char code_class;
  int lang_res;

  /* Get the next node from the queue.  */
  dq = di->queue;
  stn = dq->node;
  t = (tree) stn->key;
  dni = (dump_node_info_p) stn->value;
  index = dni->index;

  /* verify graph in sync with queue */
  if ((index - 1) != the_graph.curNodePos) {
    fprintf(stderr,"Out of sync between graph and queue (%u,%lu)\n",
	index, the_graph.curNodePos);
    exit(1);
  }

  if (index >= the_graph.nodeBuffLen) {
    if (!(gd_grow_nodes(&the_graph,DEFAULT_GROW_NODES))){
      fprintf(stderr,"Out of memory!!\n");
      exit(1);
    }
    if (index >= the_graph.nodeBuffLen) {
      /* allocation failed somehow */
      fprintf(stderr,"Error allocating nodes\n");
      exit(1);
    }
  }

  /* Remove the node from the queue, and put it on the free list.  */
  di->queue = dq->next;
  if (!di->queue)
    di->queue_end = 0;
  dq->next = di->free_list;
  di->free_list = dq;

  /* Figure out what kind of node this is.  */
  code = TREE_CODE (t);
  code_class = TREE_CODE_CLASS (code);

if(dni->binfo_p) {
  fprintf(stderr,"@%d\tbinfo",index);
} else {
  fprintf(stderr,"@%d\t%s",index,tree_code_name[(int)code]);
}

  /* some of the types are only handled in the language
   * specific version.  To make sure none slip through, we
   * initialize the node type to NODE_NONE, and check in
   * the switch default to make sure any such nodes have
   * been handled */

  the_graph.nodeBuffer[the_graph.curNodePos].common.type = NODE_NONE; 
  the_graph.nodeBuffer[the_graph.curNodePos].common.other_node = -1; 

  /**********************************
   * transfer node contents to graph
   **********************************/
  /* start with flags */
  the_graph.nodeBuffer[the_graph.curNodePos].common.unsigned_flag =
    t->common.unsigned_flag;
  the_graph.nodeBuffer[the_graph.curNodePos].common.constant_flag =
    t->common.constant_flag;
  the_graph.nodeBuffer[the_graph.curNodePos].common.static_flag =
    t->common.static_flag;
  the_graph.nodeBuffer[the_graph.curNodePos].common.public_flag =
    t->common.public_flag;
  the_graph.nodeBuffer[the_graph.curNodePos].common.private_flag =
    t->common.private_flag;
  the_graph.nodeBuffer[the_graph.curNodePos].common.protected_flag =
    t->common.protected_flag;

  /* generic per class stuff */

  if (IS_EXPR_CODE_CLASS(code_class)) {
    the_graph.nodeBuffer[the_graph.curNodePos].expr.type =
      queue_and_return_index(di,TREE_TYPE(t),DUMP_NONE)-1;
    switch (code_class) {
      case '1':
	the_graph.nodeBuffer[the_graph.curNodePos].expr1.op0 =
	  queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
	break;
      case '2':
	the_graph.nodeBuffer[the_graph.curNodePos].expr2.op0 =
	  queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
	the_graph.nodeBuffer[the_graph.curNodePos].expr2.op1 =
	  queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
	break;
      case '<':
	the_graph.nodeBuffer[the_graph.curNodePos].expr_ineq.op0 =
	  queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
	the_graph.nodeBuffer[the_graph.curNodePos].expr_ineq.op1 =
	  queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
	break;
      case 'e':
	/* handled explicitly */
	break;
      default:
	abort();
    }
  } else if (DECL_P(t)){
    the_graph.nodeBuffer[the_graph.curNodePos].decl.name=
      queue_and_return_index(di,DECL_NAME(t),DUMP_NONE)-1;
    if (DECL_ASSEMBLER_NAME(t) &&
	DECL_ASSEMBLER_NAME(t) != DECL_NAME(t)){
      the_graph.nodeBuffer[the_graph.curNodePos].decl.mngl=
	queue_and_return_index(di,DECL_ASSEMBLER_NAME(t),DUMP_NONE)-1;
    } else {
      the_graph.nodeBuffer[the_graph.curNodePos].decl.mngl= -1;
    }
    the_graph.nodeBuffer[the_graph.curNodePos].decl.type =
      queue_and_return_index(di,TREE_TYPE(t),DUMP_NONE)-1;
    the_graph.nodeBuffer[the_graph.curNodePos].decl.scpe =
      queue_and_return_index(di,DECL_CONTEXT(t),DUMP_NONE)-1;
    the_graph.nodeBuffer[the_graph.curNodePos].decl.chan =
      queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
    if (DECL_SOURCE_FILE(t)){
      int file_index;
      file_index = gd_add_string(&the_graph,DECL_SOURCE_FILE(t));
      if (file_index == -1) {
	fprintf(stderr,"error adding string %s to graph hash table\n", 
		  DECL_SOURCE_FILE(t));
	exit(1);
      }
      the_graph.nodeBuffer[the_graph.curNodePos].decl.srcf = 
	file_index;
      the_graph.nodeBuffer[the_graph.curNodePos].decl.srcl =
	DECL_SOURCE_LINE(t);
    } else {
      the_graph.nodeBuffer[the_graph.curNodePos].decl.srcf = -1;
      the_graph.nodeBuffer[the_graph.curNodePos].decl.srcl = -1;
    }
    the_graph.nodeBuffer[the_graph.curNodePos].decl.artificial =
      (DECL_ARTIFICIAL(t))?1:0;
  } else if (code_class == 't'){
    int quals = C_TYPE_QUALS(t);
    if ( quals == TYPE_UNQUALIFIED) {
      the_graph.nodeBuffer[the_graph.curNodePos].type.qual_const=0;
      the_graph.nodeBuffer[the_graph.curNodePos].type.qual_volatile=0;
      the_graph.nodeBuffer[the_graph.curNodePos].type.qual_restrict=0;
    } else {
      the_graph.nodeBuffer[the_graph.curNodePos].type.qual_const=
	(quals & TYPE_QUAL_CONST)?1:0;
      the_graph.nodeBuffer[the_graph.curNodePos].type.qual_volatile=
	(quals & TYPE_QUAL_VOLATILE)?1:0;
      the_graph.nodeBuffer[the_graph.curNodePos].type.qual_restrict=
	(quals & TYPE_QUAL_RESTRICT)?1:0;
    }
    the_graph.nodeBuffer[the_graph.curNodePos].type.name=
      queue_and_return_index(di,TYPE_NAME(t),DUMP_NONE)-1;
    the_graph.nodeBuffer[the_graph.curNodePos].type.unql=
      queue_and_return_index(di,TYPE_MAIN_VARIANT(t),DUMP_NONE)-1;
    the_graph.nodeBuffer[the_graph.curNodePos].type.size=
      queue_and_return_index(di,TYPE_SIZE(t),DUMP_NONE)-1;
    the_graph.nodeBuffer[the_graph.curNodePos].type.algn=
      TYPE_ALIGN(t);
  } else if (code_class == 'c') {
    the_graph.nodeBuffer[the_graph.curNodePos].gcc_cst.type =
      queue_and_return_index(di,TREE_TYPE(t),DUMP_NONE)-1;
  }

  /* call language specific transfer -- currently only defined for c++ */
  lang_res = 0;
  if (lang_dump_tree) {
    lang_res = (*lang_dump_tree)(di, t);
  }

fprintf(stderr,"\tAFTER C++");
  /* stuff specific to each node type */
  switch(code) {
    case NAMESPACE_DECL: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gNAMESPACE_DECL;
      break;
    } /* NAMESPACE_DECL */

    case FUNCTION_DECL: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFUNCTION_DECL;
      the_graph.nodeBuffer[the_graph.curNodePos].function_decl.args =
	queue_and_return_index(di,DECL_ARGUMENTS(t),DUMP_NONE)-1;
      if (DECL_LANG_SPECIFIC(t)) {
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.body =
	  queue_and_return_index(di,DECL_SAVED_TREE(t),DUMP_NONE)-1;
      } else {
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.body = -1;
      }
      the_graph.nodeBuffer[the_graph.curNodePos].function_decl.undef=0;
      if (DECL_EXTERNAL(t)){
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.undef=1;
      } else {
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.undef=0;
      }
      if (TREE_PUBLIC(t)){
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.ext_static=1;
      } else {
	the_graph.nodeBuffer[the_graph.curNodePos].function_decl.ext_static=0;
      }
      break;
    } /* FUNCTION_DECL */

    case VAR_DECL: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gVAR_DECL;
      the_graph.nodeBuffer[the_graph.curNodePos].var_decl.init =
	queue_and_return_index(di,DECL_INITIAL(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].var_decl.size =
	queue_and_return_index(di,DECL_SIZE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].var_decl.algn =
	DECL_ALIGN(t);
      the_graph.nodeBuffer[the_graph.curNodePos].var_decl.used =
	TREE_USED(t);
      the_graph.nodeBuffer[the_graph.curNodePos].var_decl.reg =
	(DECL_REGISTER(t))?1:0;
      break;
    } /* VAR_DECL */

    case PARM_DECL: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gPARM_DECL;
      the_graph.nodeBuffer[the_graph.curNodePos].parm_decl.argt =
	queue_and_return_index(di,DECL_ARG_TYPE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].parm_decl.size =
	queue_and_return_index(di,DECL_SIZE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].parm_decl.algn =
	DECL_ALIGN(t);
      the_graph.nodeBuffer[the_graph.curNodePos].parm_decl.used =
	TREE_USED(t);
      the_graph.nodeBuffer[the_graph.curNodePos].parm_decl.reg =
	(DECL_REGISTER(t))?1:0;
      break;
    } /* PARM_DECL */

    case TYPE_DECL: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTYPE_DECL;
      break;
    } /* TYPE_DECL */

    case LABEL_DECL: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gLABEL_DECL;
      break;
    } /* LABEL_DECL */

    case CONST_DECL: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCONST_DECL;
      the_graph.nodeBuffer[the_graph.curNodePos].const_decl.cnst =
	queue_and_return_index(di,DECL_INITIAL(t),DUMP_NONE)-1;
      break;
    } /* CONST_DECL */

    case RESULT_DECL: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gRESULT_DECL;
      the_graph.nodeBuffer[the_graph.curNodePos].result_decl.init =
	queue_and_return_index(di,DECL_INITIAL(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].result_decl.size =
	queue_and_return_index(di,DECL_SIZE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].result_decl.algn =
	DECL_ALIGN(t);
      break;
    } /* RESULT_DECL */

    case FIELD_DECL: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFIELD_DECL;
      the_graph.nodeBuffer[the_graph.curNodePos].field_decl.init =
	queue_and_return_index(di,DECL_INITIAL(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].field_decl.size =
	queue_and_return_index(di,DECL_SIZE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].field_decl.algn =
	DECL_ALIGN(t);
      the_graph.nodeBuffer[the_graph.curNodePos].field_decl.bitfield =
	DECL_C_BIT_FIELD(t);
      if (DECL_FIELD_OFFSET(t)){
	the_graph.nodeBuffer[the_graph.curNodePos].field_decl.bpos =
	  queue_and_return_index(di,bit_position(t),DUMP_NONE)-1;
      } else {
	the_graph.nodeBuffer[the_graph.curNodePos].field_decl.bpos = -1;
      }
      break;
    } /* FIELD_DECL */

    /* template_decl  - C++ only */
    /* using_decl  - C++ only */

    case FUNCTION_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFUNCTION_TYPE;
      the_graph.nodeBuffer[the_graph.curNodePos].function_type.retn=
	queue_and_return_index(di,TREE_TYPE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].function_type.prms=
	queue_and_return_index(di,TYPE_ARG_TYPES(t),DUMP_NONE)-1;
      break;
    } /* FUNCTION_TYPE */

    case INTEGER_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gINTEGER_TYPE;
      the_graph.nodeBuffer[the_graph.curNodePos].integer_type.prec=
	TYPE_PRECISION(t);
      the_graph.nodeBuffer[the_graph.curNodePos].integer_type.unsgnd=
	TREE_UNSIGNED(t);
      the_graph.nodeBuffer[the_graph.curNodePos].integer_type.min=
	queue_and_return_index(di,TYPE_MIN_VALUE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].integer_type.max=
	queue_and_return_index(di,TYPE_MAX_VALUE(t),DUMP_NONE)-1;
      break;
    } /* INTEGER_TYPE */

    case ARRAY_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gARRAY_TYPE;
      the_graph.nodeBuffer[the_graph.curNodePos].array_type.elts=
	queue_and_return_index(di,TREE_TYPE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].array_type.domn=
	queue_and_return_index(di,TYPE_DOMAIN(t),DUMP_NONE)-1;
      break;
    } /* ARRAY_TYPE */

    case VOID_TYPE:
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gVOID_TYPE;
      break;

    case REAL_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gREAL_TYPE;
      the_graph.nodeBuffer[the_graph.curNodePos].real_type.prec=
	TYPE_PRECISION(t);
      break;
    } /* REAL_TYPE */

    case COMPLEX_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCOMPLEX_TYPE;
      break;
    } /* COMPLEX_TYPE */

    case VECTOR_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gVECTOR_TYPE;
      break;
    } /* VECTOR_TYPE */

    case ENUMERAL_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gENUMERAL_TYPE;
      the_graph.nodeBuffer[the_graph.curNodePos].enumeral_type.prec=
	TYPE_PRECISION(t);
      the_graph.nodeBuffer[the_graph.curNodePos].enumeral_type.unsgnd=
	TREE_UNSIGNED(t);
      the_graph.nodeBuffer[the_graph.curNodePos].enumeral_type.min=
	queue_and_return_index(di,TYPE_MIN_VALUE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].enumeral_type.max=
	queue_and_return_index(di,TYPE_MAX_VALUE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].enumeral_type.csts=
	queue_and_return_index(di,TYPE_VALUES(t),DUMP_NONE)-1;
      break;
    } /* ENUMERAL_TYPE */

    case BOOLEAN_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gBOOLEAN_TYPE;
      break;
    } /* BOOLEAN_TYPE */

    case CHAR_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCHAR_TYPE;
      break;
    } /* CHAR_TYPE */

    case POINTER_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gPOINTER_TYPE;
      if (!lang_res) {
	the_graph.nodeBuffer[the_graph.curNodePos].pointer_type.ptd =
	  queue_and_return_index(di,TREE_TYPE(t),DUMP_NONE)-1;
	the_graph.nodeBuffer[the_graph.curNodePos].pointer_type.ptrmem = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].pointer_type.cls = -1;
      }
      break;
    } /* POINTER_TYPE */

    case OFFSET_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gOFFSET_TYPE;
      break;
    } /* OFFSET_TYPE */

    case REFERENCE_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gREFERENCE_TYPE;
      the_graph.nodeBuffer[the_graph.curNodePos].reference_type.refd =
	queue_and_return_index(di,TREE_TYPE(t),DUMP_NONE)-1;
      break;
    } /* REFERENCE_TYPE */

    case METHOD_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gMETHOD_TYPE;
      the_graph.nodeBuffer[the_graph.curNodePos].method_type.clas =
	queue_and_return_index(di,TYPE_METHOD_BASETYPE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].method_type.retn=
	queue_and_return_index(di,TREE_TYPE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].method_type.prms=
	queue_and_return_index(di,TYPE_ARG_TYPES(t),DUMP_NONE)-1;
      break;
    } /* METHOD_TYPE */

    case FILE_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFILE_TYPE;
      break;
    } /* FILE_TYPE */

    case SET_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSET_TYPE;
      break;
    } /* SET_TYPE */

    case RECORD_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gRECORD_TYPE;
      the_graph.nodeBuffer[the_graph.curNodePos].record_type.flds =
	queue_and_return_index(di,TYPE_FIELDS(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].record_type.fncs =
	queue_and_return_index(di,TYPE_METHODS(t),DUMP_NONE)-1;
      if (lang_res == 0) {
	the_graph.nodeBuffer[the_graph.curNodePos].record_type.ptrmem = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].record_type.isClass = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].record_type.binf = -1;
      }
      break;
    } /* RECORD_TYPE */

    case UNION_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gUNION_TYPE;
      the_graph.nodeBuffer[the_graph.curNodePos].union_type.flds =
	queue_and_return_index(di,TYPE_FIELDS(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].union_type.fncs =
	queue_and_return_index(di,TYPE_METHODS(t),DUMP_NONE)-1;
      if (lang_res == 0) {
	the_graph.nodeBuffer[the_graph.curNodePos].union_type.ptrmem = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].union_type.binf = -1;
      }
      break;
    } /* UNION_TYPE */

    case QUAL_UNION_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gQUAL_UNION_TYPE;
      break;
    } /* QUAL_UNION_TYPE */

    case LANG_TYPE: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gLANG_TYPE;
      break;
    } /* LANG_TYPE */

    /* template_type_parm  -- C++ only */
    /* template_parm_index  -- C++ only */
    /* template_template_parm  -- C++ only */
    /* bound_template_template_parm  -- C++ only */
    /* typename_type  -- C++ only */
    /* typeof_type  -- C++ only */

    case BLOCK: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gBLOCK;
      break;
    } /* BLOCK */

    case COMPOUND_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCOMPOUND_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].compound_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].compound_stmt.body =
	queue_and_return_index(di,COMPOUND_BODY(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].compound_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* COMPOUND_STMT */

    case EXPR_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gEXPR_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].expr_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].expr_stmt.texpr =
	queue_and_return_index(di,EXPR_STMT_EXPR(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].expr_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* EXPR_STMT */

    case DECL_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gDECL_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].decl_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].decl_stmt.decl =
	queue_and_return_index(di,DECL_STMT_DECL(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].decl_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* DECL_STMT */

    case SCOPE_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSCOPE_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].scope_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].scope_stmt.begin_end =
	(!SCOPE_BEGIN_P(t))?1:0;
      the_graph.nodeBuffer[the_graph.curNodePos].scope_stmt.null =
	(SCOPE_NULLIFIED_P(t))?1:0;
      the_graph.nodeBuffer[the_graph.curNodePos].scope_stmt.clnp =
	(!SCOPE_NO_CLEANUPS_P(t))?1:0;
      the_graph.nodeBuffer[the_graph.curNodePos].scope_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* SCOPE_STMT */

    case RETURN_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gRETURN_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].return_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].return_stmt.ret_expr =
	queue_and_return_index(di,RETURN_EXPR(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].return_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* RETURN_STMT */

    case STMT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSTMT_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].stmt_expr.stmt = 
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* STMT_EXPR */

    case BREAK_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gBREAK_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].break_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].break_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* BREAK_STMT */

    case CONTINUE_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCONTINUE_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].continue_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].continue_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* CONTINUE_STMT */

    case IF_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gIF_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].if_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].if_stmt.cond =
	queue_and_return_index(di,IF_COND(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].if_stmt.thenc =
	queue_and_return_index(di,THEN_CLAUSE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].if_stmt.elsec =
	queue_and_return_index(di,ELSE_CLAUSE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].if_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* IF_STMT */

    case SWITCH_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSWITCH_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].switch_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].switch_stmt.cond =
	queue_and_return_index(di,SWITCH_COND(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].switch_stmt.body =
	queue_and_return_index(di,SWITCH_BODY(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].switch_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* SWITCH_STMT */

    case LABEL_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gLABEL_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].label_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].label_stmt.label =
	queue_and_return_index(di,LABEL_STMT_LABEL(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].label_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* LABEL_STMT */

    /* try_block, handler,start_catch_stmt,cleanup_stmt -- C++ only */

    case CASE_LABEL: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCASE_LABEL;
      the_graph.nodeBuffer[the_graph.curNodePos].case_label.low =
	queue_and_return_index(di,CASE_LOW(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].case_label.high =
	queue_and_return_index(di,CASE_HIGH(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].case_label.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* CASE_LABEL */

    case GOTO_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gGOTO_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].goto_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].goto_stmt.dest =
	queue_and_return_index(di,GOTO_DESTINATION(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].goto_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* GOTO_STMT */

    case FOR_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFOR_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].for_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].for_stmt.init =
	queue_and_return_index(di,FOR_INIT_STMT(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].for_stmt.cond =
	queue_and_return_index(di,FOR_COND(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].for_stmt.inc_expr =
	queue_and_return_index(di,FOR_EXPR(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].for_stmt.body =
	queue_and_return_index(di,FOR_BODY(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].for_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* FOR_STMT */

    case WHILE_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gWHILE_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].while_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].while_stmt.cond =
	queue_and_return_index(di,WHILE_COND(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].while_stmt.body =
	queue_and_return_index(di,WHILE_BODY(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].while_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* WHILE_STMT */

    case DO_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gDO_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].do_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].do_stmt.body =
	queue_and_return_index(di,DO_BODY(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].do_stmt.cond =
	queue_and_return_index(di,DO_COND(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].do_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* DO_STMT */

    case ASM_STMT: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gASM_STMT;
      the_graph.nodeBuffer[the_graph.curNodePos].asm_stmt.line =
	STMT_LINENO(t);
      the_graph.nodeBuffer[the_graph.curNodePos].asm_stmt.vol =
	(ASM_VOLATILE_P(t))?1:0;
      the_graph.nodeBuffer[the_graph.curNodePos].asm_stmt.strg =
	queue_and_return_index(di,ASM_STRING(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].asm_stmt.outs =
	queue_and_return_index(di,ASM_OUTPUTS(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].asm_stmt.ins =
	queue_and_return_index(di,ASM_INPUTS(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].asm_stmt.clbr =
	queue_and_return_index(di,ASM_CLOBBERS(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].asm_stmt.next =
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* ASM_STMT */

    case COMPONENT_REF: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCOMPONENT_REF;
      the_graph.nodeBuffer[the_graph.curNodePos].component_ref.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].component_ref.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* COMPONENT_REF */

    /* bit_field_ref */
    case BIT_FIELD_REF: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gBIT_FIELD_REF;
      break;
    } /* BIT_FIELD_REF */

    case INDIRECT_REF: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gINDIRECT_REF;
      the_graph.nodeBuffer[the_graph.curNodePos].indirect_ref.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      break;
    } /* INDIRECT_REF */

    /* buffer_ref */

    case ARRAY_REF: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gARRAY_REF;
      the_graph.nodeBuffer[the_graph.curNodePos].array_ref.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].array_ref.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* ARRAY_REF */

    /* offset_ref scope_ref member_ref - C++ only*/

    case CONSTRUCTOR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCONSTRUCTOR;
      the_graph.nodeBuffer[the_graph.curNodePos].constructor.elts = 
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* CONSTRUCTOR */

    /* compound_expr  */
    case COMPOUND_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCOMPOUND_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].compound_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].compound_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* COMPOUND_EXPR */

    case MODIFY_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gMODIFY_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].modify_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].modify_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* MODIFY_EXPR */

    case INIT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gINIT_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].init_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].init_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* INIT_EXPR */

    /* throw_expr -- C++ only */
    case TARGET_EXPR: {
      node_ptr init;
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTARGET_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].target_expr.decl =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].target_expr.init =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].target_expr.clnp =
	queue_and_return_index(di,TREE_OPERAND(t,2),DUMP_NONE)-1;
      init =  queue_and_return_index(di,TREE_OPERAND(t,3),DUMP_NONE)-1;
      if (the_graph.nodeBuffer[the_graph.curNodePos].target_expr.init == -1){
	the_graph.nodeBuffer[the_graph.curNodePos].target_expr.init = init;
      } else {
	if (init != -1){
	  fprintf(stderr,"comment in c-dump.c about TARGET_EXPR.init is false!!\n");
	}
	if (init == the_graph.nodeBuffer[the_graph.curNodePos].target_expr.init){
	  fprintf(stderr,"however, they both point to the same entity\n");
	}
      }
      break;
    } /* TARGET_EXPR */

    case COND_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCOND_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].cond_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].cond_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].cond_expr.op2 =
	queue_and_return_index(di,TREE_OPERAND(t,2),DUMP_NONE)-1;
      break;
    } /* COND_EXPR */

    case BIND_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gBIND_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].bind_expr.vars =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].bind_expr.body =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* BIND_EXPR */

    case CALL_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCALL_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].call_expr.fn =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].call_expr.args =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* CALL_EXPR */

    case METHOD_CALL_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gMETHOD_CALL_EXPR;
      break;
    } /* METHOD_CALL_EXPR */

    case WITH_CLEANUP_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gWITH_CLEANUP_EXPR;
      break;
    } /* WITH_CLEANUP_EXPR */

    case PLACEHOLDER_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gPLACEHOLDER_EXPR;
      break;
    } /* PLACEHOLDER_EXPR */

    case WITH_RECORD_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gWITH_RECORD_EXPR;
      break;
    } /* WITH_RECORD_EXPR */

    case CLEANUP_POINT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCLEANUP_POINT_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].cleanup_point_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      break;
    } /* CLEANUP_POINT_EXPR */

    case PLUS_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gPLUS_EXPR;
      break;
    } /* PLUS_EXPR */

    case MINUS_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gMINUS_EXPR;
      break;
    } /* MINUS_EXPR */

    case MULT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gMULT_EXPR;
      break;
    } /* MULT_EXPR */

    case TRUNC_DIV_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTRUNC_DIV_EXPR;
      break;
    } /* TRUNC_DIV_EXPR */

    case CEIL_DIV_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCEIL_DIV_EXPR;
      break;
    } /* CEIL_DIV_EXPR */

    case FLOOR_DIV_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFLOOR_DIV_EXPR;
      break;
    } /* FLOOR_DIV_EXPR */

    case ROUND_DIV_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gROUND_DIV_EXPR;
      break;
    } /* ROUND_DIV_EXPR */

    case TRUNC_MOD_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTRUNC_MOD_EXPR;
      break;
    } /* TRUNC_MOD_EXPR */

    case CEIL_MOD_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCEIL_MOD_EXPR;
      break;
    } /* CEIL_MOD_EXPR */

    case FLOOR_MOD_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFLOOR_MOD_EXPR;
      break;
    } /* FLOOR_MOD_EXPR */

    case ROUND_MOD_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gROUND_MOD_EXPR;
      break;
    } /* ROUND_MOD_EXPR */

    case RDIV_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gRDIV_EXPR;
      break;
    } /* RDIV_EXPR */

    case EXACT_DIV_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gEXACT_DIV_EXPR;
      break;
    } /* EXACT_DIV_EXPR */

    case FIX_TRUNC_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFIX_TRUNC_EXPR;
      break;
    } /* FIX_TRUNC_EXPR */

    case FIX_CEIL_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFIX_CEIL_EXPR;
      break;
    } /* FIX_CEIL_EXPR */

    case FIX_FLOOR_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFIX_FLOOR_EXPR;
      break;
    } /* FIX_FLOOR_EXPR */

    case FIX_ROUND_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFIX_ROUND_EXPR;
      break;
    } /* FIX_ROUND_EXPR */

    case FLOAT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFLOAT_EXPR;
      break;
    } /* FLOAT_EXPR */

    case EXPON_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gEXPON_EXPR;
      break;
    } /* EXPON_EXPR */

    case NEGATE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gNEGATE_EXPR;
      break;
    } /* NEGATE_EXPR */

    case MIN_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gMIN_EXPR;
      break;
    } /* MIN_EXPR */

    case MAX_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gMAX_EXPR;
      break;
    } /* MAX_EXPR */

    case ABS_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gABS_EXPR;
      break;
    } /* ABS_EXPR */

    case FFS_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gFFS_EXPR;
      break;
    } /* FFS_EXPR */

    case LSHIFT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gLSHIFT_EXPR;
      break;
    } /* LSHIFT_EXPR */

    case RSHIFT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gRSHIFT_EXPR;
      break;
    } /* RSHIFT_EXPR */

    case LROTATE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gLROTATE_EXPR;
      break;
    } /* LROTATE_EXPR */

    case RROTATE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gRROTATE_EXPR;
      break;
    } /* RROTATE_EXPR */

    case BIT_IOR_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gBIT_IOR_EXPR;
      break;
    } /* BIT_IOR_EXPR */

    case BIT_XOR_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gBIT_XOR_EXPR;
      break;
    } /* BIT_XOR_EXPR */

    case BIT_AND_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gBIT_AND_EXPR;
      break;
    } /* BIT_AND_EXPR */

    case BIT_ANDTC_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gBIT_ANDTC_EXPR;
      break;
    } /* BIT_ANDTC_EXPR */

    case BIT_NOT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gBIT_NOT_EXPR;
      break;
    } /* BIT_NOT_EXPR */

    case TRUTH_ANDIF_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTRUTH_ANDIF_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].truth_andif_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].truth_andif_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* TRUTH_ANDIF_EXPR */

    case TRUTH_ORIF_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTRUTH_ORIF_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].truth_orif_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].truth_orif_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* TRUTH_ORIF_EXPR */

    case TRUTH_AND_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTRUTH_AND_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].truth_and_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].truth_and_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* TRUTH_AND_EXPR */

    case TRUTH_OR_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTRUTH_OR_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].truth_or_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].truth_or_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* TRUTH_XOR_EXPR */

    case TRUTH_XOR_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTRUTH_XOR_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].truth_xor_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].truth_xor_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* TRUTH_XOR_EXPR */

    case TRUTH_NOT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTRUTH_NOT_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].truth_not_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      break;
    } /* TRUTH_NOT_EXPR */

    case LT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gLT_EXPR;
      break;
    } /* LT_EXPR */

    case LE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gLE_EXPR;
      break;
    } /* LE_EXPR */

    case GT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gGT_EXPR;
      break;
    } /* GT_EXPR */

    case GE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gGE_EXPR;
      break;
    } /* GE_EXPR */

    case EQ_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gEQ_EXPR;
      break;
    } /* EQ_EXPR */

    case NE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gNE_EXPR;
      break;
    } /* NE_EXPR */

    case UNORDERED_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gUNORDERED_EXPR;
      break;
    } /* UNORDERED_EXPR */

    case ORDERED_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gORDERED_EXPR;
      break;
    } /* ORDERED_EXPR */

    case UNLT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gUNLT_EXPR;
      break;
    } /* UNLT_EXPR */

    case UNLE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gUNLE_EXPR;
      break;
    } /* UNLE_EXPR */

    case UNGT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gUNGT_EXPR;
      break;
    } /* UNGT_EXPR */

    case UNGE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gUNGE_EXPR;
      break;
    } /* UNGE_EXPR */

    case UNEQ_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gUNEQ_EXPR;
      break;
    } /* UNEQ_EXPR */

    case IN_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gIN_EXPR;
      break;
    } /* IN_EXPR */

    case SET_LE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSET_LE_EXPR;
      break;
    } /* SET_LE_EXPR */

    case CARD_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCARD_EXPR;
      break;
    } /* CARD_EXPR */

    case RANGE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gRANGE_EXPR;
      break;
    } /* RANGE_EXPR */

    case CONVERT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCONVERT_EXPR;
      break;
    } /* CONVERT_EXPR */

    case NOP_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gNOP_EXPR;
      break;
    } /* NOP_EXPR */

    case NON_LVALUE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gNON_LVALUE_EXPR;
      break;
    } /* NON_LVALUE_EXPR */

    case SAVE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSAVE_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].save_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      break;
    } /* SAVE_EXPR */

    case UNSAVE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gUNSAVE_EXPR;
      break;
    } /* UNSAVE_EXPR */

    case RTL_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gRTL_EXPR;
      break;
    } /* RTL_EXPR */

    case ADDR_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gADDR_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].addr_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      break;
    } /* ADDR_EXPR */

    case REFERENCE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gREFERENCE_EXPR;
      break;
    } /* REFERENCE_EXPR */

    case ENTRY_VALUE_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gENTRY_VALUE_EXPR;
      break;
    } /* ENTRY_VALUE_EXPR */

    case COMPLEX_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCOMPLEX_EXPR;
      break;
    } /* COMPLEX_EXPR */

    case CONJ_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCONJ_EXPR;
      break;
    } /* CONJ_EXPR */

    case REALPART_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gREALPART_EXPR;
      break;
    } /* REALPART_EXPR */

    case PREDECREMENT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gPREDECREMENT_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].predecrement_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].predecrement_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* PREDECREMENT_EXPR */

    case PREINCREMENT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gPREINCREMENT_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].preincrement_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].preincrement_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* PREINCREMENT_EXPR */

    case POSTDECREMENT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gPOSTDECREMENT_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].postdecrement_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].postdecrement_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* POSTDECREMENT_EXPR */

    case POSTINCREMENT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gPOSTINCREMENT_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].postincrement_expr.op0 =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].postincrement_expr.op1 =
	queue_and_return_index(di,TREE_OPERAND(t,1),DUMP_NONE)-1;
      break;
    } /* POSTINCREMENT_EXPR */

    case VA_ARG_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gVA_ARG_EXPR;
      break;
    } /* VA_ARG_EXPR */

    case TRY_CATCH_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTRY_CATCH_EXPR;
      break;
    } /* TRY_CATCH_EXPR */

    case TRY_FINALLY_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTRY_FINALLY_EXPR;
      break;
    } /* TRY_FINALLY_EXPR */

    /* goto_subroutine -- C++ only */

    case POPDHC_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gPOPDHC_EXPR;
      break;
    } /* POPDHC_EXPR */

    case POPDCC_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gPOPDCC_EXPR;
      break;
    } /* POPDCC_EXPR */

    case LABEL_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gLABEL_EXPR;
      break;
    } /* LABEL_EXPR */

    case GOTO_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gGOTO_EXPR;
      break;
    } /* GOTO_EXPR */

    case RETURN_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gRETURN_EXPR;
      break;
    } /* RETURN_EXPR */

    case EXIT_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gEXIT_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].exit_expr.cond =
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      break;
    } /* EXIT_EXPR */

    case LOOP_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gLOOP_EXPR;
      the_graph.nodeBuffer[the_graph.curNodePos].loop_expr.body=
	queue_and_return_index(di,TREE_OPERAND(t,0),DUMP_NONE)-1;
      break;
    } /* LOOP_EXPR */

    case LABELED_BLOCK_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gLABELED_BLOCK_EXPR;
      break;
    } /* LABELED_BLOCK_EXPR */

    case EXPR_WITH_FILE_LOCATION: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gEXPR_WITH_FILE_LOCATION;
      the_graph.nodeBuffer[the_graph.curNodePos].expr_with_file_location.expr2=
	queue_and_return_index(di,EXPR_WFL_NODE(t),DUMP_NONE)-1;
      break;
    } /* EXPR_WITH_FILE_LOCATION */

    case SWITCH_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSWITCH_EXPR;
      break;
    } /* SWITCH_EXPR */

    case SIZEOF_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSIZEOF_EXPR;
      break;
    } /* SIZEOF_EXPR */

    case ARROW_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gARROW_EXPR;
      break;
    } /* ARROW_EXPR */

    case ALIGNOF_EXPR: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gALIGNOF_EXPR;
      break;
    } /* ALIGNOF_EXPR */

    /* vec_nw_expr, nw_expr vec_dl_expr dl_expr
     * type_expr aggr_init_expr empty_class_expr template_id_expr
     * lookup_expr function_name subobject ctor_stmt ctor_initializer
     * return_init tag_defn identity_conv lvalue_conv qual_conv std_conv
     * ptr_conv pmem_conv base_conv, ref_bind, user_conv, ambig_conv, rvalue_conv
     * -- C++ only */

    case INTEGER_CST: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gINTEGER_CST;
      the_graph.nodeBuffer[the_graph.curNodePos].integer_cst.high =
	TREE_INT_CST_HIGH(t);
      the_graph.nodeBuffer[the_graph.curNodePos].integer_cst.low =
	TREE_INT_CST_LOW(t);
      break;
    } /* INTEGER_CST */

    case STRING_CST: {
      int string_idx;
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gSTRING_CST;
      string_idx = gd_add_string(&the_graph,TREE_STRING_POINTER(t));
      if (string_idx == -1){
	fprintf(stderr,"error adding string %s to graph hash table\n",
	    	TREE_STRING_POINTER(t));
        exit(1);
      }
      the_graph.nodeBuffer[the_graph.curNodePos].string_cst.strg =
	string_idx;
      the_graph.nodeBuffer[the_graph.curNodePos].string_cst.lngt =
	TREE_STRING_LENGTH(t);
      break;
    } /* STRING_CST */

    case COMPLEX_CST: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gCOMPLEX_CST;
      break;
    } /* COMPLEX_CST */

    case REAL_CST: {
      REAL_VALUE_TYPE d;
      char string[100];
      str_ptr string_idx;

      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gREAL_CST;

      /* adapted from tree-print.c */

#if !defined(REAL_IS_NOT_DOUBLE) || defined(REAL_ARITHMETIC)

      d = TREE_REAL_CST(t);
      if (REAL_VALUE_ISINF(d)){
	string_idx = gd_add_string(&the_graph,"Inf");
	if (string_idx == -1) {
	  fprintf(stderr,"error adding string Inf to graph hash table\n");
	  exit(1);
	}
      } if (REAL_VALUE_ISNAN(d)){
	string_idx = gd_add_string(&the_graph,"NaN");
	if (string_idx == -1) {
	  fprintf(stderr,"error adding string NaN to graph hash table\n");
	  exit(1);
	}
      } else {
	REAL_VALUE_TO_DECIMAL(d,"%e",string);
	string_idx = gd_add_string(&the_graph,string);
	if (string_idx == -1) {
	  fprintf(stderr,"error adding string %s to graph hash table\n",string);
	  exit(1);
	}
      }

#else

      {
	int i;
	unsigned char * p = (unsigned char *) &TREE_REAL_CST(t);
	unsigned char * d = (unsgiend char *) string;
	*d++ = '0';
	*d++ = 'x';
	for(i = 0; i < sizeof(TREE_REAL_CST(t)); i++){
	  unsigned char hexDigits[3];
	  sprintf(hexDigits,"%02x",*p++);
	  *d++ = hexDigits[0];
	  *d++ = hexDigits[1];
	}
	*d = '\0';
	/* string is now null terminated hex value of real_constant */
	string_idx = gd_add_string(&the_graph,string);
	if (string_idx == -1) {
	  fprintf(stderr,"error adding string %s to graph hash table\n",string);
	  exit(1);
	}
      }

#endif
      the_graph.nodeBuffer[the_graph.curNodePos].real_cst.value = string_idx;

      break;
    } /* REAL_CST */

    /* ptrmem_cst
     * -- C++ only */

    case OP_IDENTIFIER: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gOP_IDENTIFIER;
      break;
    } /* OP_IDENTIFIER */

    case IDENTIFIER_NODE: {
      int str_index;
      /* should check lang_res */
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gIDENTIFIER_NODE;
      if (IDENTIFIER_POINTER(t)){
	/* printf("Identifier node at %d = %s\n",index,IDENTIFIER_POINTER(t)); */
	str_index = gd_add_string(&the_graph,IDENTIFIER_POINTER(t));
	if(str_index == -1) {
	    fprintf(stderr,"error adding string %s to graph hash table", 
		      IDENTIFIER_POINTER(t));
	    exit(1);
	}
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.strg = str_index;
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.lngt = IDENTIFIER_LENGTH(t);
      } else {
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.strg = -1;
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.lngt = 0;
      }
      if (lang_res == 0) {
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.op_type_un = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].identifier_node.tynm = -1;
      }
      break;
    } /* IDENTIFIER_NODE */

    case TREE_LIST: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTREE_LIST;
      the_graph.nodeBuffer[the_graph.curNodePos].tree_list.purpose = 
	queue_and_return_index(di,TREE_PURPOSE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].tree_list.value = 
	queue_and_return_index(di,TREE_VALUE(t),DUMP_NONE)-1;
      the_graph.nodeBuffer[the_graph.curNodePos].tree_list.chan = 
	queue_and_return_index(di,TREE_CHAIN(t),DUMP_NONE)-1;
      break;
    } /* TREE_LIST */

    case TREE_VEC: {
      /* only interested in binfos */
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gTREE_VEC;
      if (dni->binfo_p) {
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.binfo = 1;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.inherit = 
	  queue_and_return_index(di,BINFO_INHERITANCE_CHAIN(t),DUMP_BINFO)-1;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.pub =
	  TREE_VIA_PUBLIC(t)?1:0;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.prot =
	  TREE_VIA_PROTECTED(t)?1:0;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.priv =
	  TREE_VIA_PRIVATE(t)?1:0;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.virt =
	  TREE_VIA_VIRTUAL(t)?1:0;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.type =
	  queue_and_return_index(di,BINFO_TYPE(t),DUMP_BINFO)-1;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.base =
	  queue_and_return_index(di,BINFO_BASETYPES(t),DUMP_BINFO)-1;
      } else {
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.binfo = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.inherit = -1;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.pub = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.prot = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.priv = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.virt = 0;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.type = -1;
	the_graph.nodeBuffer[the_graph.curNodePos].tree_vec.base = -1;
      }
      break;
    } /* TREE_VEC */

    case ERROR_MARK: {
      the_graph.nodeBuffer[the_graph.curNodePos].common.type = gERROR_MARK;
      break;
    } /* ERROR_MARK */

    /* default_arg, binding, overload, wrapper
     * -- C++ only */


    default:
      if (the_graph.nodeBuffer[the_graph.curNodePos].common.type == NODE_NONE){
	fprintf(stderr,"Unkown node type %d\n",code);
	exit(1);
      }
  }
/* fprintf(stderr,"foobar\n"); */
  /* increment number of nodes in graph */
  the_graph.curNodePos++;

fprintf(stderr,"\n");

  return;
}

/* Dump T, and all its children, on STREAM.  */

static void
dump_node (t, stream)
     tree t;
     FILE *stream;
{
  struct dump_info di;
  dump_queue_p dq;
  dump_queue_p next_dq;

  /* Initialize the dump-information structure.  */

  di.stream = stream;
  di.index = 0;
  di.column = 0;
  di.queue = 0;
  di.queue_end = 0;
  di.free_list = 0;
  di.nodes = splay_tree_new (splay_tree_compare_pointers, 0, 
			     (splay_tree_delete_value_fn) &free);

  /* Queue up the first node.  */
  queue (&di, t, DUMP_NONE);

  /* Until the queue is empty, keep dumping nodes.  */
  while (di.queue)
    dequeue_and_dump (&di);

  /* Now, clean up.  */
  for (dq = di.free_list; dq; dq = next_dq)
    {
      next_dq = dq->next;
      free (dq);
    }
  splay_tree_delete (di.nodes);
}

/* Dump T, and all its children, to FILE.  */

void
dump_node_to_file (t, file)
     tree t;
     const char *file;
{

  /* need either arguments on the command line, or
   * we need someway to predict sizes. For now, hard code
   * the sizes -- T.D. */

  the_graph = allocate_gd_graph(16* 8192, 16 * 4096, 16 *2048);

  dump_node(t,(FILE*)stdout); 

  the_graph.firstNode = 0; /* first node of a gcc dump is always the
			      first node in the buffer */

  gd_graph_write_binary_file(file,the_graph);
}
